#!/usr/bin/perl
#wakeOnLAN.pl

use English;

# flush after every write
$| = 1;

use constant false => 0;
use constant true  => 1;

use strict;
#use warnings;
use POSIX qw(strftime);
use Sys::Hostname;
use IO::Socket::INET6;
use NetAddr::IP;
use Tie::File;
use Time::Local;

my ($in_Socket,$out_Socket,$Magic_Packet);
my ($Packet_Length,$Magic_Packet_Hex,$Valid_Magic_Packet);
my ($RemoteAddress,$MACAddress,$Password,$Computer);
my ($LastRemoteAddress,$LastMACAddress,$LastValid_Magic_Packet,$LastTimeStamp);

sub Signal_Handler
{
	$in_Socket->close();
	`synologset1 sys info 0x11B01051 "Wake on LAN Gateway"`;
	Log_Message("Wake on LAN Gateway service was stopped.");
	unlink "./WakeOnLAN.pid";
	exit 0;
} 
$SIG{KILL} = \&Signal_Handler;
$SIG{TERM} = \&Signal_Handler;
$SIG{INT} = \&Signal_Handler;
$SIG{QUIT} = \&Signal_Handler;

sub Log_Message
{
	if (($_[0]) ne "")
	{
		my $lGo = true;
		if (($_[1]) == true)
		{
			my $TimeStamp = timegm(localtime);
			my $Seconds = $TimeStamp - $LastTimeStamp;
			if (($LastRemoteAddress eq $RemoteAddress) and ($LastMACAddress eq $MACAddress) and ($Seconds < 11))
			{
				if ($Valid_Magic_Packet == $LastValid_Magic_Packet)
				{
					$lGo = false;
				}
				else
				{
					if ($LastValid_Magic_Packet == true)
					{
						$lGo = false;
					}
					else
					{
						tie my @Log, 'Tie::File' , '/var/packages/WakeOnLAN/target/gateway/WOL-Gateway.log';
						$#Log -=1;
						untie @Log;
						$LastValid_Magic_Packet = $Valid_Magic_Packet;
					}
				}
			}
			else
			{
				$LastRemoteAddress = $RemoteAddress;
				$LastMACAddress = $MACAddress;
				$LastValid_Magic_Packet = $Valid_Magic_Packet;
			}
			$LastTimeStamp = timegm(localtime);
			if ($lGo == true)
			{
				if ($LastValid_Magic_Packet == true) 
				{
					`synologset1 sys info 0x11100000 "Wake on LAN accepted from $RemoteAddress for $Computer"`;
				}
				else
				{
					`synologset1 sys warn 0x11100000 "Wake on LAN rejected from $RemoteAddress for $Computer"`;
				}
			}
		}
		else
		{
			$LastRemoteAddress = "";
			$LastMACAddress = "";
			$LastValid_Magic_Packet = false;
			$LastTimeStamp = timegm(localtime);
		}
		if ($lGo == true)
		{
			my $datestring = strftime "%Y %b %e (%a) %H:%M:%S", localtime;
			if (($_[2]) == true)
			{
				open(my $logfile, '>', "/var/packages/WakeOnLAN/target/gateway/WOL-Gateway.log") or die "Could not open the logfile $!";
				print $logfile "$datestring $_[0]\n";
				close $logfile;
			}
			else
			{
				open(my $logfile, '>>', "/var/packages/WakeOnLAN/target/gateway/WOL-Gateway.log") or die "Could not open the logfile $!";
				print $logfile "$datestring $_[0]\n";
				close $logfile;
			}
		}
	}
}

sub Check_Local_Address
{
	my $IP_Local = false;
	my $FilterAddress = ($_[0]);
	my $Interfaces = `ifconfig | grep "Bcast:"`;
	chomp($Interfaces);
	my @Interfaces = split('\n', $Interfaces);
	foreach my $Interface (@Interfaces)
	{
		$Interface =~ s/^\s+//;
		$Interface =~ s/  /:/g;
		my @values = split(':', $Interface);
		my $count = 0;
		foreach my $value (@values)
		{
			if ($count == 1 && $value eq $FilterAddress)
			{
				$IP_Local = true;
			}
			$count = $count + 1;
		} 
	}
	return $IP_Local;	
}

sub Check_IP_Address
{
	my $IP_OK = false;
	my $FilterAddress = ($_[0]);
	if (($RemoteAddress eq $FilterAddress) || ($FilterAddress eq "") or ($FilterAddress eq "0.0.0.0"))
	{
		$IP_OK = true;
	}
	else
	{
		my $FilterNET = "";
		my $RemoteNET = "";
		$FilterNET = NetAddr::IP->new($FilterAddress);
		$RemoteNET = NetAddr::IP->new($RemoteAddress);
		if ($FilterNET->contains($RemoteNET))
		{
			$IP_OK = true;
		}
	}
	return $IP_OK;
}

sub Filter_Magic_Packet
{
	$Computer = $MACAddress;
	my $Filter_OK = true;
	if ( -e '/var/packages/WakeOnLAN/target/gateway/WOL-Gateway.conf' )
	{
		my $Filter_Active = false;
		$Filter_OK = false;
		open my $filter, '/var/packages/WakeOnLAN/target/gateway/WOL-Gateway.conf' or die "Could not open configuration: $!";
		while (( my $row = <$filter> ) && ($Filter_OK == false))
		{
			chomp($row);
			if (substr($row, 0, 1) ne "#")
			{
				$Filter_Active = true;
				my $Filter_Computer = "";
				my $Filter_MACAddress = "";
				my $Filter_RemoteAddress = "";
				my $Filter_Password = "";
				my $Filter_PassOn = "";
				my @values = split(',', $row, 5);
				my $count = 0;
				foreach my $value (@values)
				{

					if ( $count == 4 )

					{
						$Filter_PassOn = lc($value);
						
					}

					else

					{

						if ( $count == 3 )

						{
							if ( length($value) > 0 )
							{
								if ( length($value) > 6 )
								{ 
									$Filter_Password = lc($value);
								}
								else
								{
									$Filter_Password = unpack("H*", $value);
									while ( length($Filter_Password) < 12 )
									{
										$Filter_Password = $Filter_Password . "ff";
									}
								}
							}
						}

						else

						{
							if ( $count == 2 )
							
							{
							
								$Filter_RemoteAddress = $value;
								
							}

							else

							{
								if ( $count == 1 )
								
								{
								
									$Filter_MACAddress = lc($value);
									
								}
								else
								
								{
								
									$Filter_Computer = $value;
									
								}

							}

						}
					}
					$count = $count + 1;
				}
				if ($Filter_MACAddress eq $MACAddress)
				{
					$Computer = $Filter_Computer;
					if ((Check_IP_Address($Filter_RemoteAddress) == true) && ($Filter_Password eq $Password))
					{	
						$Filter_OK = true;
						if (($Filter_PassOn ne "true") and (length($Magic_Packet) == 108))
						{
							$Magic_Packet = pack("H*", substr($Magic_Packet_Hex, 0, 204));
						}
					}
				}
			}
		}
		close $filter;
		if ($Filter_Active == false)
		{
			$Filter_OK = true;
		}
	}
	return $Filter_OK;
}

# Write the Process ID
open(my $pidfile, '>', "/var/packages/WakeOnLAN/target/gateway/WOL-Gateway.pid") or die "Could not open the pid file $!";
print $pidfile "$PROCESS_ID";
close $pidfile;

# Create the socket on UDP port 9 waiting for Wake on LAN Magic packets
$in_Socket = new IO::Socket::INET (
LocalPort => '19999',
Proto => 'udp',
) or die "Couldn't create the socket : $!\n";

`synologset1 sys info 0x11B01050 "Wake on LAN Gateway"`;
Log_Message("Wake on LAN Gateway service was started.", false, true);
while(1)
{
	# Wait for Wake on LAN Magic packet
	$in_Socket->recv($Magic_Packet,1024);

	if ((length($Magic_Packet) == 102) || (length($Magic_Packet) == 108))
	{
		#Get the IP address of the sender of the Magic packet
		$RemoteAddress = $in_Socket->peerhost();
		if (Check_Local_Address($RemoteAddress) == false) 
		{
			$Valid_Magic_Packet = false;
			
			#Validate the Magic packet
			$Magic_Packet_Hex = unpack("H*", $Magic_Packet);

			if (substr($Magic_Packet_Hex, 0, 12) eq 'ffffffffffff')
			{
				if ((substr($Magic_Packet_Hex, 12, 12) eq substr($Magic_Packet_Hex, 24, 12))
				&&  (substr($Magic_Packet_Hex, 12, 12) eq substr($Magic_Packet_Hex, 36, 12))
				&&  (substr($Magic_Packet_Hex, 12, 12) eq substr($Magic_Packet_Hex, 48, 12))
				&&  (substr($Magic_Packet_Hex, 12, 12) eq substr($Magic_Packet_Hex, 60, 12))
				&&  (substr($Magic_Packet_Hex, 12, 12) eq substr($Magic_Packet_Hex, 72, 12))
				&&  (substr($Magic_Packet_Hex, 12, 12) eq substr($Magic_Packet_Hex, 84, 12))
				&&  (substr($Magic_Packet_Hex, 12, 12) eq substr($Magic_Packet_Hex, 96, 12))
				&&  (substr($Magic_Packet_Hex, 12, 12) eq substr($Magic_Packet_Hex, 108, 12))
				&&  (substr($Magic_Packet_Hex, 12, 12) eq substr($Magic_Packet_Hex, 120, 12))
				&&  (substr($Magic_Packet_Hex, 12, 12) eq substr($Magic_Packet_Hex, 132, 12))
				&&  (substr($Magic_Packet_Hex, 12, 12) eq substr($Magic_Packet_Hex, 144, 12))
				&&  (substr($Magic_Packet_Hex, 12, 12) eq substr($Magic_Packet_Hex, 156, 12))
				&&  (substr($Magic_Packet_Hex, 12, 12) eq substr($Magic_Packet_Hex, 168, 12))
				&&  (substr($Magic_Packet_Hex, 12, 12) eq substr($Magic_Packet_Hex, 180, 12))
				&&  (substr($Magic_Packet_Hex, 12, 12) eq substr($Magic_Packet_Hex, 192, 12)))
				{
					$MACAddress = substr($Magic_Packet_Hex, 12, 2) . ":" . substr($Magic_Packet_Hex, 14, 2) . ":" . substr($Magic_Packet_Hex, 16, 2) . ":" . substr($Magic_Packet_Hex, 18, 2) . ":" . substr($Magic_Packet_Hex, 20, 2) . ":" . substr($Magic_Packet_Hex, 22, 2);
					if (length($Magic_Packet) == 108)
					{
						$Password = substr($Magic_Packet_Hex, 204, 2) . substr($Magic_Packet_Hex, 206, 2) . substr($Magic_Packet_Hex, 208, 2) . substr($Magic_Packet_Hex, 210, 2) . substr($Magic_Packet_Hex, 212, 2) . substr($Magic_Packet_Hex, 214, 2);
					}
					else
					{
						$Password = "";
					}
					$Valid_Magic_Packet = Filter_Magic_Packet();
					if ($Valid_Magic_Packet == true)
					{
						Log_Message("Wake on LAN accepted from $RemoteAddress for $Computer.", true);
						#Create the socket to broadcast the Magic Packet
						my $Interfaces = `ifconfig | grep "Bcast:"`;
						chomp($Interfaces);
						my @Interfaces = split('\n', $Interfaces);
						foreach my $Interface (@Interfaces)
						{
							my $Gateway = "";
							$Interface =~ s/^\s+//;
							$Interface =~ s/  /:/g;
							my @values = split(':', $Interface);
							my $count = 0;
							foreach my $value (@values)
							{
								if ($count == 3)
								{	
									$Gateway = $value;
								}
								$count = $count + 1;
							}
							$out_Socket = new IO::Socket::INET (
							PeerAddr => $Gateway,
							PeerPort => '9',
							Proto => 'udp',
							Broadcast => 1,
							) or die "Couldn't create the socket : $!\n";
							$out_Socket->send($Magic_Packet);
							$out_Socket->close();
						}
					}
					else
					{
						Log_Message("Wake on LAN rejected from $RemoteAddress for $Computer.", true);
					}
				}
				else
				{
					Log_Message("Invalid packet received from $RemoteAddress!");
				}
			}
			else
			{
				Log_Message("Invalid packet received from $RemoteAddress!");
			}
		}
	}
}
$in_Socket->close();
